	class cATTParser extends cATTParserBase
//	---------------------------------------
{
//	Parser error messages

	static	s_errstr = ["",
		"Invalid or ambiguous keyword",			// INVCMD
		"Syntax error",					// SYNTAX
		"Invalid use of qualifier or parameter",	// INVUSE
		"Missing or invalid value",			// INVVAL
		"Missing or invalid parameter",			// INVPARM
		"Invalid qualifier",				// INVQUAL
		"Too many parameters",				// TOOMANY
		"Required parameter missing",			// PARMREQ
		"Conflicting qualifiers",			// CONFLICT
		"Internal error detected by parser",		// PARSER
		"Unmatched parentheses",			// UNMATCH
		];

//	Parser regular expressions

	static	s_rexkeyword	= @"^([ \t]*)(\a+)($|[ \t/])";
	static	s_rexqualifier	= @"^([ \t]*/)(\a+)($|[ \t/:])";
	static	s_rexint	= @"^([ \t]*)(-?\d+)($|[ \t/,])";
	static	s_rexsymbol	= @"^([ \t]*)([_\a]\w*)($|[ \t/,=])";
	static	s_rexsymbolw	= @"^([ \t]*)([_\a*%]+[\w*%]*)($|[ \t/,])";
	static	s_rexeol	= @"^([ \t]*)(.*)";
	static	s_rexvar	= @"^([ \t]*)([\a_].*)($|[=,])";
	static	s_rexscope	= @"^([ \t]*)((?:(?:\d+)(?:$|\s))|(?:[\a_]\w*))"
	static	s_rexlinerange	=
		@"^([ \t]*)(?:(?:([\a_]\w*)\\)?(\d*)?(?::(\d+))?){1}($|[ \t/])";
	static	s_rexaddr	=
		@"^([ \t]*)((?:[\a_]\w*)\\)?((?:\d+)|(?:[\a_]\w*))($|[ \t/,])";
	static	s_rexwatch	=
		@"^([ \t]*)(?:((?:[\a_]\w*)\\)?([\a_].*)){1}($|[ \t/,])";

	c_parse = null;					// Command syntax
	c_exkeyword = null;				// Compiled rex
	c_exqualifier = null;
	c_exsymbol = null;

	static	s_rexarray = [	null,				// 1+
				"s_rexaddr",			// _ADDR
				"s_rexint",			// _INT
				null,				// _STR
				"s_rexsymbol",			// _SYMBOL
				"s_rexwatch",			// _WATCH
				null,				// _MODULE
				"s_rexscope",			// _SCOPE
				"s_rexlinerange",		// _LINERANGE
				"s_rexvar",			// _VAR
				null,				// _FILE
				"s_rexeol",			// _EOL
				"s_rexsymbolw",			// _SYMBOLW
			    ];

	c_definet = null;				// Defines table

	c_cmd = null;					// Command line
	c_cmdlen = null;				// Command line length

	c_cmdt = null;					// Command table
	c_cmda = null;					// Command array
	c_qualt = null;					// Qualifier table
	c_parma = null;					// Parameter array

	c_ret = null;					// Parser return table
}
	function cATTParser::Initialize(t)
//	----------------------------------
//	Set up the parsing rules table
//	c_parse is a table that describes the command formats.
//	Each entry is a primary keyword,primary table pair.
//	The primary keyword is the first word of the command
//	The primary table contains the second keyword or _NULL_ and the
//	primary array as the table value
//	The primary array has the following format:
//		[command id,qualifiers table]
//
//	The qualifiers table has a qualifier or parameter constant paired
//	with a qualifier description array.
//	The qualifier description array has the following format:
//		[qualifier name,qualifier value type] or
//		[_Pn_,parameter description array] or
//		[_CONFLICT_,[[qual,qual],...]
//	The qualifier value type can be:
//		null	for value not allowed or
//		array	to describe the value allowed or
//	The qualifier value array has the following format:
//		[negated flag, value type, min value, max value]
//	If a negated flag is specified, it can be as follows:
//		_HASNO	has negated form, ie: /SILENT has /NOSILENT
//		_ISNO	is the negated form, ie: /NOSILENT
//	The allowed qualifier value types are:
//		_INT	for an integer value
//
//	If the qualifier name is of the form _Pn_, it describes parameter n
//	which can be between 1 and 9.
//	The parameter description array has the following format:
//		[parameter type,parameter count] or
//		[_INT, count, min value, max value]
//	The parameter type can be: as defined in cattparser_base or string
//		ex: _ADDR or
//		string		a string value to match exactly
//	The parameter count can be:
//		Between 0 and 9, specifies exact number of parameters required
//		_ZERO_PLUS specifies zero to 9 optional parameters
//		_ONE_PLUS specifies at least one, but up to 9 parameters
//
//	If the qualifier name is _CONFLICT_, the next array element is an
//	array of arrays, with each sub-array listing the qualifiers that
//	cannot be specified together or an ATT_PE_CONFLICT error will result.
{
	c_parse = t;

	c_exkeyword = regexp(s_rexkeyword);		// Compile rex
	c_exqualifier = regexp(s_rexqualifier);
	c_exsymbol = regexp(s_rexsymbol);

	c_definet = {	d = "deposit",			// Single char shortcut
			e = "examine",
			s = "step",
		    };
}
	function cATTParser::ParseCommand(cmd)
//	--------------------------------------
//	Parse a command line according to the rules in c_parse
//	Returns	Execution table with following fields:
//		cmd		Original command line
//		errcode		ATT_PE_ code
//		errpos		Position of error in cmd
//		newcmd		If defined symbol found (only if)(not parsed)
//		funcid		Function id to be executed
//		kword1		First keyword
//		kword2		Second keyword
//		qualt		Qualifier table with array values
//
//		The qualt table has the following format:
//			qualifier name, qaulifier array or parameter array
//		The qaulifier array has the following format:
//			[negated,value,...]
//		The parameter array has only parameter values
{
	local	kwrec,kwrec2,errcode;

	c_cmd = ::strip(cmd);
	c_cmdlen = c_cmd.len();

	c_ret = {};
	c_ret.cmd <- c_cmd;				// Command line

	while (true) {
	  if(c_cmd[0] == '@') {
	    c_cmd = "_AT " + c_cmd.slice(1);
	    c_cmdlen = c_cmd.len();
	  }

	  kwrec = ParseKeyword(c_exsymbol,c_cmd,0,c_parse);
	  if(kwrec.errcode == ATT_PE_SUCCESS) break;

	  if(kwrec.errcode != ATT_PE_INVCMD) return Error(kwrec.errcode,0);

	  errcode = CheckDefines(kwrec.kword,kwrec.cmdpos);
	  if(errcode != ATT_PE_SUCCESS) return Error(errcode,0);

	  c_ret.newcmd <- c_cmd;
	  c_ret.errcode <- ATT_PE_SUCCESS;
	  return c_ret;					// Just substituted
	}

	c_ret.kword1 <- kwrec.kword;			// 1st keyword

	c_cmdt = c_parse[kwrec.kword];			// Commands table

//	If there is a second keyword, find and match it
//	If c_cmdt table has _NULL_, there are no second keywords for this cmd

	if("_NULL_" in c_cmdt) {			// Just one keyword
	  c_ret.kword2 <- null;				// No 2nd keyword

	  c_cmda = c_cmdt._NULL_;			// Command array
	}
	else {						// Has 2nd keyword
	  if(kwrec.eol) return Error(ATT_PE_INVCMD,0);

	  local kwrec2 = ParseKeyword(c_exkeyword,c_cmd,kwrec.cmdpos,c_cmdt);

	  if(kwrec2.errcode != ATT_PE_SUCCESS)
	    return Error(kwrec2.errcode,kwrec.cmdpos);

	  c_ret.kword2 <- kwrec2.kword;			// 2nd keyword

	  c_cmda = c_cmdt[kwrec2.kword];		// Command array

	  kwrec = clone kwrec2;
	}

	c_ret.funcid <- c_cmda[I_FUNCID];		// Cmd function id
	c_ret.qualt <- {};				// Qualifier table

//	If the command has qualifiers or parameters, parse them here

	if(c_qualt = c_cmda[I_QUALIFIERS]) {		// Has table
	  local	rec;

	  if(kwrec.eol) {				// At end of line
	    rec = TestForParameters(kwrec.cmdpos);
	    if(rec.errcode < 0) return rec;
	  }
	  else {
	    while (!kwrec.eol) {
	      rec = TestForQualifier(kwrec.cmdpos);
	      if(rec.errcode < 0) return rec;

	      if(!rec.errcode) {
		rec = TestForParKeyword("WHEN",kwrec.cmdpos);
	        if(rec.errcode < 0) return rec;
	      }

	      if(!rec.errcode) {
		rec = TestForParKeyword("DO",kwrec.cmdpos);
	        if(rec.errcode < 0) return rec;
	      }

	      if(!rec.errcode) {
	        rec = TestForParameters(kwrec.cmdpos);
	        if(rec.errcode < 0) return rec;

		if(!rec.errcode) return Error(ATT_PE_TOOMANY,kwrec.cmdpos);
	      }

	      kwrec = clone rec;
	    }						// While !eol
	  }
	}
	else if(!kwrec.eol)
	  return Error(ATT_PE_INVUSE,kwrec.cmdpos);

	c_ret.errcode <- ATT_PE_SUCCESS;
	return c_ret;					// Exec table
}
	function cATTParser::ParseKeyword(exkeyword,cmd,cmdpos,cmdtab)
//	--------------------------------------------------------------
//	Attempt to parse and match a keyword.
//	exkeyword	Compiled regex to use to find keyword
//	cmd		Command line
//	cmdpos		Position to start parsing
//	cmdtab		Command table to search for keyword
//	Returns		A table with the following fields
//			  cmdpos	pos for next token
//			  eol		true if at end-of-line
//			  kword		keyword found if not syntax error
//			  errcode	ATT_PE_ code
//			  delimiter	stopped scanning on this char or null
{
	local	ct,kword,ret={};

	ct = exkeyword.capture(cmd,cmdpos);		// Get cmd keyword
	if(!ct) {
	  ret.errcode <- ATT_PE_SYNTAX;
	  return ret;
	}

	ret.cmdpos <- ct[2].end;			// Next token start
	ret.eol <- (ct[2].end == cmd.len());

	ret.kword <- cmd.slice(ct[2].begin,ct[2].end);
	ret.delimiter <- (ct[3].end) ? cmd.slice(ct[3].begin,ct[3].end) : null;

	if(kword = FindKeyword(cmdtab,ret.kword)) {
	  ret.kword <- kword;
	  ret.errcode <- ATT_PE_SUCCESS;
	}
	else
	  ret.errcode <- ATT_PE_INVCMD;

	return ret;
}
	function cATTParser::FindKeyword(t,token)
//	-----------------------------------------
//	Scan table t for unambiguous match of token
//	Returns null if not found or is ambiguous, or the full keyword name in t
{
	if(!t) return null;

	local	match = null;
	local	toklen = token.len();
	local	utoken = token.toupper();

	foreach (kw,v in t) {
	  if(kw.len() >= toklen && kw.slice(0,toklen) == utoken) {
	    if(match) return null;			// Ambiguous

	    match = kw;
	  }
	}

	return match;
}
	function cATTParser::CheckDefines(kword,cmdpos)
//	-----------------------------------------------
//	Check if the keyword is defined in c_definet, make substitution if so
//	Updates c_cmd and c_cmdlen
//	kword		keyword to find in c_definet
//	cmdpos		Rest of command in c_cmd
//	Returns		ATT_PE_ code
{
	local	ukword = kword.toupper();

	foreach (def,str in c_definet) {
	  if(def.toupper() == ukword) {			// Match
	    local eol = (cmdpos == c_cmdlen);
	    local subst=false;
	    local cmd="",strlen = str.len();
	    local parms = SplitParameters(c_cmd.slice(cmdpos));

	    for (local i=0; i < strlen; i++) {
	      if(str[i] == '#' || str[i] == '&') {
		if(i+1 == strlen) return ATT_PE_SYNTAX;

		local p = str[i+1] - '0';		// Parm number

		if(p < 1 || p > 9) return ATT_PE_SYNTAX;

		if(parms.len() >= p)
		  cmd += parms[p-1];
		else if(str[i] == '#')			// Must have
		  return ATT_PE_PARMREQ;

		i++;					// Skip parm number
		subst = true;				// Found subst params
	      }
	      else
		cmd += str[i].tochar();			// Move from equnam
	    }

	    if(!subst && !eol) cmd += c_cmd.slice(cmdpos);

	    c_cmd = strip(cmd);
	    c_cmdlen = c_cmd.len();
	    return ATT_PE_SUCCESS;
	  }
	}

	return ATT_PE_INVCMD;
}
	function cATTParser::DefineSymbol(name,value)
//	---------------------------------------------
//	Define a symbolic name and its value
{
	c_definet[name] <- value;
}
	function cATTParser::DeleteSymbol(name)
//	---------------------------------------
//	Delete a defined symbolic name
//	name	A symbol name or $ALL to delete all symbols
{
	if(name == "$ALL")
	  c_definet = {};
	else
	  delete c_definet[name];
}
	function cATTParser::GetSymbolTable()
//	-------------------------------------
//	Returns the defines table, so it can be searched and listed
{
	return c_definet;
}
	function cATTParser::Error(errcode,errpos=0)
//	--------------------------------------------
//	Create and return the parser return table. Contains just a few fields.
{
	local t = {cmd = c_cmd, errcode = errcode, errpos = errpos};

	return t;
}
	function cATTParser::GetErrorStr(errcode)
//	-----------------------------------------
//	Return error string for error code
{
	return s_errstr[-errcode];
}
	function cATTParser::TestForQualifier(cmdpos)
//	---------------------------------------------
//	Returns	A table with the following fields:
//		errcode		If 0, no qualifier found (only field)
//				If ATT_PE_SUCCESS, fields returned by
//				ParseKeyword or ParseQualifier
//				Otherwise fields returned by Error
{
	local	qrec = ParseKeyword(c_exqualifier,c_cmd,cmdpos,c_qualt);

	switch (qrec.errcode) {
	case ATT_PE_SUCCESS:
	  return ParseQualifier(qrec,cmdpos);

	case ATT_PE_SYNTAX:
	  return {errcode = 0};				// Not qualifier

	case ATT_PE_INVCMD:
	  return Error(ATT_PE_INVQUAL,cmdpos);
	}

	return qrec;
}
	function cATTParser::ParseQualifier(qrec,cmdpos)
//	------------------------------------------------
//	Add qualifier to c_ret.qualt table
//	If qualifier is negatable or negated, remove other form from table
//	qrec		Table returned by ParseKeyword
//	cmdpos		Position of qualifier before calling ParseKeyword
//	Returns		Table with errcode, cmdpos, eol
{
	local quala = c_qualt[qrec.kword];		// Qualifier array

	c_ret.qualt[qrec.kword] <- null;		// Define qualifier

	if(quala) {					// Has array value
	  if(quala[I_NEGFLG] >= _HASNO) {		// Has negated flag
	    local	qual;

	    switch (quala[I_NEGFLG]) {
	    case _HASNO:				// Has NO form
	      qual = "NO" + qrec.kword;
	      break;

	    case _ISNO:					// Is the NO form
	      qual = qrec.kword.slice(2);		// Without NO
	      break;
	    }

	    if(qual in c_ret.qualt)			// Other form in c_ret table
	      delete c_ret.qualt[qual];			// Delete it
	  }
	  else {					// Has value array
	    if(qrec.delimiter != ":") return Error(ATT_PE_INVVAL,qrec.cmdpos);

	    local pval = TestForValues(qrec.kword,qrec.cmdpos+1);

	    if(!pval.errcode) return Error(ATT_PE_INVVAL,pval.cmdpos);
	    if(pval.errcode != ATT_PE_SUCCESS) return pval;

	    qrec.cmdpos = pval.cmdpos;
	  }
	}

//	Check for conflicting qualifiers

	local conf = TestConflict(cmdpos);

	if(conf.errcode != ATT_PE_SUCCESS) return conf;

	local	ret = {};

	ret.errcode <- ATT_PE_SUCCESS;
	ret.cmdpos <- qrec.cmdpos;
	ret.eol <- (qrec.cmdpos == c_cmdlen);

	return ret;
}
	function cATTParser::TestConflict(cmdpos)
//	-----------------------------------------
//	Check for conflicting qualifiers
//	Returns		Error table
{
	if("_CONFLICT_" in c_qualt) {			// _CONFLICT_
	  local	confa = c_qualt._CONFLICT_;		// Conflict array

	  foreach (i,arr in confa) {
	    local count = 0;				// Qualifier count

	    foreach (j,cqual in arr)			// Conflicting qualifiers
	      if(cqual in c_ret.qualt) count++;		// Qualifier defined

	    if(count > 1) return Error(ATT_PE_CONFLICT,cmdpos);
	  }
	}

	return {errcode = ATT_PE_SUCCESS};
}
	function cATTParser::TestForParameters(cmdpos)
//	----------------------------------------------
//	Check if _Pn_ exists in c_qualt and process if so
//	Returns		A table with the following fields
//			  errcode	ATT_PE_ code or
//					0 if no parms (only field)
//			  cmdpos	pos for next token
//			  eol		true if at end-of-line
{
	local	i,ret={},pval,iparm=1;

	for (i=1; i <= 9; i++) {
	  local name = "_P" + (i + '0').tochar() + "_";

	  if(!(name in c_qualt)) {			// Parameter not defined
	    if(i == iparm) return {errcode = 0};
	    break;
	  }

//	  Check if parameter name is already defined in parser result
//	  table, if so, skip to next one. Update iparm to next number.

	  if(name in c_ret.qualt) {			// Already parsed
	    iparm++;					// Look for next
	    continue;
	  }

	  pval = TestForValues(name,cmdpos);

	  if(!pval.errcode) {
	    if(i == 1) return {errcode = 0};
	    break;
	  }

	  if(pval.errcode != ATT_PE_SUCCESS) return pval;

	  cmdpos = pval.cmdpos;
	}

	ret.errcode <- ATT_PE_SUCCESS;
	ret.cmdpos <- cmdpos;
	ret.eol <- (cmdpos == c_cmdlen);

	return ret;
}
	function cATTParser::TestForValues(name,cmdpos)
//	-----------------------------------------------
//	Test for parameter or qualifier values
//	name		Parameter or qualifier name
//	cmdpos		Start parsing here
//	Returns		Table with errcode and cmdpos
//			errcode is 0 if no values needed nor found
{
	local	pval;

	c_parma = c_qualt[name];			// Parameter array

	if(cmdpos == c_cmdlen) {			// At EOL
	  if(c_parma[I_COUNT] <= _ONE_PLUS)		// Need at least 1
	    return Error(ATT_PE_INVPARM,cmdpos);

	  return {errcode = 0};
	}

//	Look for expected parameter type

	switch (::type(c_parma[I_PARMTYPE])) {
	case "integer":
	  pval = ParseValues(c_parma[I_PARMTYPE],cmdpos);
	  break;

	case "string":
	  pval = ParseLiteral(cmdpos);
	  break;

	default:
	  return Error(ATT_PE_PARSER,cmdpos);
	}

	if(pval.errcode != ATT_PE_SUCCESS)
	  return Error(pval.errcode,pval.cmdpos);

//	Value count

	if(c_parma[I_PARMTYPE] != _LINERANGE) {
	  if((c_parma[I_COUNT] == _ZERO_OR_ONE && pval.values.len() > 1) ||
	     (c_parma[I_COUNT] <= 9 && pval.values.len() > c_parma[I_COUNT]))
	    return Error(ATT_PE_TOOMANY,cmdpos);

	  else if(c_parma[I_COUNT] <= 9 && pval.values.len() < c_parma[I_COUNT])
	    return Error(ATT_PE_PARMREQ,cmdpos);
	}

	c_ret.qualt[name] <- pval.values;		// Parameter values

	local	conf = TestConflict(cmdpos);

	if(conf.errcode != ATT_PE_SUCCESS) return conf;

	return {errcode = ATT_PE_SUCCESS, cmdpos = pval.cmdpos};
}
	function cATTParser::ParseValues(parmtype,cmdpos)
//	-------------------------------------------------
//	Get parameter values, if _INT, check for valid range
//	parmtype	Parser parameter type, ex: _INT
//	cmdpos		Position to start parsing at
//	Returns		A table with the following fields
//			  errcode	ATT_PE_ code
//			  cmdpos	pos for next token
//			  eol		true if at end-of-line
//			  values	Array of values
{
	local	regex,ex,ct,pos,ival,ret={};

	if(s_rexarray[parmtype] == null) {
	  AILog.Error("Parameter type " + parmtype + " not in s_rexarray");
	  return {errcode = ATT_PE_PARSER, cmdpos = cmdpos};
	}

	regex = this[s_rexarray[parmtype]];		// Regular expression
	ex = regexp(regex);

	ret.values <- [];

	while (true) {					// Max 9 values
	  ct = ex.capture(c_cmd,cmdpos);
	  if(!ct) {
	    ret.errcode <- ATT_PE_INVVAL;
	    break;
	  }

//	  Post processing

	  if(parmtype == _INT) {			// Verify value range
	    ival = c_cmd.slice(ct[2].begin,ct[2].end).tointeger();

	    if(ival < c_parma[I_MINVAL] || ival > c_parma[I_MAXVAL]) {
	      ret.errcode <- ATT_PE_INVVAL;
	      break;
	    }

	    ret.values.push(ival);
	    cmdpos = ct[2].end;				// Next token start
	  }
	  else if(parmtype == _ADDR || parmtype == _WATCH) {
	    if(ct[2].end > ct[2].begin)			// Module
	      ret.values.push(c_cmd.slice(ct[2].begin,ct[2].end-1));
	    else
	      ret.values.push(null);			// No module

	    local v2 = c_cmd.slice(ct[3].begin,ct[3].end);

	    if(v2[0] <= '9') {				// Line number
	      v2 = v2.tointeger();
	      if(v2 < 1) {
	        ret.errcode <- ATT_PE_INVVAL;
	        break;
	      }
	    }

	    ret.values.push(v2);			// Line or function name
	    cmdpos = ct[3].end;				// Next token start
	  }
	  else if(parmtype == _LINERANGE) {	// Break into parts: module,start,end
	    if(ct[2].end > ct[2].begin)	{		// Module
	      ret.values.push(c_cmd.slice(ct[2].begin,ct[2].end));
	      cmdpos = ct[2].end;			// Next token start
	    }
	    else
	      ret.values.push(null);			// No module

	    local ival1 = null;

	    if(ct[3].begin < ct[3].end) {		// Start line
	      ival1 = c_cmd.slice(ct[3].begin,ct[3].end).tointeger();

	      if(ival1 < 1) {
	        ret.errcode <- ATT_PE_INVVAL;
	        break;
	      }

	      ret.values.push(ival1);			// start
	      cmdpos = ct[3].end;			// Next token start
	    }
	    else
	      ret.values.push(null);			// No start line

	    if(ct[4].begin < ct[4].end) {		// End line
	      local ival2 = c_cmd.slice(ct[4].begin,ct[4].end).tointeger();

	      if(ival2 < 1 || (ival1 != null && ival2 < ival1)) {
	        ret.errcode <- ATT_PE_INVVAL;
	        break;
	      }

	      ret.values.push(ival2);			// end
	      cmdpos = ct[4].end;			// Next token start
	    }
	    else
	      ret.values.push(null);			// No end line
	  }
	  else {
	    ret.values.push(strip(c_cmd.slice(ct[2].begin,ct[2].end)));
	    cmdpos = ct[2].end;				// Next token start
	  }

	  if(!(pos = MoreValues(cmdpos))) {
	    ret.errcode <- ATT_PE_SUCCESS;
	    break;
	  }

	  cmdpos = pos;
	}

	ret.cmdpos <- cmdpos;
	ret.eol <- (cmdpos == c_cmdlen);

	return ret;
}
	function cATTParser::ParseLiteral(cmdpos)
//	-----------------------------------------
//	Parser is expecting a literal string, may be preceeded by blanks
//	cmdpos		Position to start parsing at
//	Returns		A table with the following fields
//			  errcode	ATT_PE_ code
//			  cmdpos	pos for next token
//			  eol		true if at end-of-line
//			  values	Array, single value, the literal
{
	local	ret = {errcode = ATT_PE_SYNTAX};

	while (cmdpos < c_cmdlen) {
	  while (cmdpos < c_cmdlen) {			// Skip whitespace
	    local c = c_cmd[cmdpos];

	    if(c != ' ' && c != '\t') break;

	    cmdpos++;
	  }

	  if(cmdpos == c_cmdlen) break;

	  local	str = c_parma[I_PARMTYPE];		// Literal to match

	  if(cmdpos + str.len() > c_cmdlen) break;

	  if(str != c_cmd.slice(cmdpos,cmdpos+str.len())) break;

	  ret.errcode = ATT_PE_SUCCESS;
	  ret.values <- [];
	  ret.values.push(str);

	  cmdpos += str.len();
	  break;
	}

	ret.cmdpos <- cmdpos;
	ret.eol <- (cmdpos == c_cmdlen);

	return ret;
}
	function cATTParser::MoreValues(cmdpos)
//	---------------------------------------
//	Check if cmdpos is followed by a comma to separate values
//	Returns		Position of next value or 0 if no comma
{
	if((cmdpos = SkipSpaces(cmdpos)) == c_cmdlen || c_cmd[cmdpos] != ',')
	  return 0;

	return cmdpos + 1;
}
	function cATTParser::SkipSpaces(cmdpos)
//	---------------------------------------
{
	while (cmdpos < c_cmdlen) {
	  switch (c_cmd[cmdpos]) {
	  case ' ':
	  case '\t':
	    break;

	  default:
	    return cmdpos;
	  }

	  cmdpos++;
	}

	return cmdpos;					// EOL
}
	function cATTParser::ParseParenthesis(cmdpos)
//	---------------------------------------------
//	Scan c_cmd for a string between ()
//	Returns		a table with the following fields:
//			  errcode	ATT_PE_ code (0 if no '(')
//			  cmdpos	pos for next token
//			  eol		true if at end-of-line
//			  value		string found, without ()
{
	local	pos=SkipSpaces(cmdpos);

	if(pos == c_cmdlen || c_cmd[pos] != '(') return Error(0);

	local	found = false;
	local	inpar = 1;				// Within '()'
	local	st = ++pos;

	while (pos < c_cmdlen && !found) {		// Find ')'
	  switch (c_cmd[pos]) {
	  case ')':
	    if(--inpar < 0) return Error(ATT_PE_UNMATCH,cmdpos);

	    if(!inpar) found = true;

	    break;

	  case '(':
	    inpar++;
	    break;

	  case '"':
	  case '\'':
	    local i = pos++;

	    while (pos < c_cmdlen && c_cmd[pos] != c_cmd[i])
	      pos++;

	    if(pos == c_cmdlen) return Error(ATT_PE_SYNTAX,cmdpos);

	    break;
	  }

	  pos++;
	}

	if(!found) return Error(ATT_PE_UNMATCH,cmdpos);

	local	ret = {};

	ret.errcode <- ATT_PE_SUCCESS;
	ret.cmdpos <- pos;
	ret.eol <- (pos == c_cmdlen);
	ret.value <- strip(c_cmd.slice(st,pos-1));

	return ret;
}
	function cATTParser::TestForParKeyword(kword,cmdpos)
//	----------------------------------------------------
//	Check if kword is followed by parenthesised value
//	Returns		table returned by ParseParenthesis
{
	if(!("_"+kword+"_" in c_qualt)) return Error(0);

	local	pos=SkipSpaces(cmdpos);
	local	en = pos + kword.len();

	if(en > c_cmdlen || c_cmd.slice(pos,en).toupper() != kword)
	  return Error(0);

	local pret = ParseParenthesis(en);

	if(pret.errcode != ATT_PE_SUCCESS) return Error(pret.errcode,cmdpos);

	c_ret.qualt["_"+kword+"_"] <- pret.value;

	return pret;
}
	function cATTParser::SplitParameters(cmdline)
//	---------------------------------------------
//	Split command into parts at ' ' outside of quotes
//	Returns		array with individual parameters or ATT_PE_ code
{
	local	st=-1, pos=0, cmdlen=cmdline.len();
	local	arr = [];

	while (pos < cmdlen) {			// Find ' '
	  switch (cmdline[pos]) {
	  case '"':
	  case '\'':
	    local i = pos++;

	    while (pos < cmdlen && cmdline[pos] != cmdline[i])
	      pos++;

	    if(pos == cmdlen) return ATT_PE_SYNTAX;

	    break;

	  case ' ':
	  case '\t':
	    if(st != -1) {			// Skip leading spaces
	      arr.push(cmdline.slice(st,pos));
	      st = -1;
	    }
	    break;

	  default:
	    if(st == -1) st = pos;		// First non-space
	  }

	  pos++;
	}

	if(st >= 0 && pos > st) arr.push(cmdline.slice(st,pos));

	return arr;
}
