	class cATTRoad extends cATTBase
//	-------------------------------
{
	c_classname = null;
	c_path = null;				// Path found
	c_pathn = null;				// Other end of path
	c_from = null;				// Starting points
	c_to = null;				// Goals
	c_ignore = null;			// Array of tiles to ignore

	c_ends = null;				// End tiles found
	c_front = null;				// Station front tiles
	c_cost = null;				// Test build cost
	c_dist = null;				// Total distance
	c_error = null;

	constructor (from,to,ignore=null)
//	---------------------------------
//	from		An array of [tile,dir] pairs
//			tile	Possible start tile
//			dir	DIR_M_ code, restricts direction from
//				start tile. If DIR_M_UNKNOWN, no restriction.
//	to		An array of tiles or [tile,prev_tile] pairs
//			tile		Possible goal tile
//			prev_tile	Restricts path thru this tile to tile
//	ignore		An array of tiles to ignore

	{
	  c_classname = "cATTRoad";
	  c_from = from;
	  c_to = to;
	  c_ignore = ignore;
	  c_cost = 0;
	  c_dist = 0;
	  c_ends = [];
	  c_front = [];
	}
}
	function cATTRoad::FindPath(maxlen=null,iterations=100)
//	-------------------------------------------------------
//	maxlen		Maximum length of a route
//	iterations	Run for iterations, then call CheckGameTasks
//			-1 to run until solution found or abandoned
{
	local	pf = cATTRoadPF();

	pf.Initialize(c_from,c_to,c_ignore);

//	If no maxlen specified, make sure the route does not detour
//	more than 20% of the best route length (20% each way = 40%)

	if(maxlen)
	  pf.SetMaxRoadLength(maxlen);

	else {
	  local goal = (typeof c_to[0] == "integer") ? c_to[0] : c_to[0][0];
	  local dist = AITile.GetDistanceManhattanToTile(c_from[0][0],goal);

	  pf.SetMaxDetour(goal,(dist * 1.2).tointeger());
	}

//	Let AStar run for a day, then execute CheckGameTasks

	local	start = AIDate.GetCurrentDate();

	while (true) {
	  c_path = pf.FindPath(iterations);

	  if(c_path == null || c_path) break;

	  local date = AIDate.GetCurrentDate();

	  if(date > start) {
	    start = date;

	    local exec = AIExecMode();
	    s_g.att.CheckDailyTasks();
	    exec = null;

	    if(s_g.abort) return null;			// Aborted
	  }
	}

	return c_path;
}
	function cATTRoad::TestBuildRoad()
//	----------------------------------
//	Test builds the road to find real cost and the other end
//	Inserts "next" fields so path can be followed from other end
//	Inserts "exists" field to speed-up real construction
//	Converts nested path tables to array of tables, if successful
//	Returns		true if road can be built
{
	local	dist,npath;
	local	acc = AIAccounting();
	local	test = AITestMode();
	local	path = c_path;
	local	built = false;

	c_dist = 0;					// Total distance
	c_ends.push(path.tile);				// Station tile
	c_front.push(path.prev.tile);			// Station front tile
	path.next <- null;				// Other dir

	while (path != null) {
	  local ppath = path.prev;

	  if(ppath == null) {
	    c_ends.push(path.tile);			// Other end tile
	    c_front.push(npath.tile);			// Tile before other end

	    path.exists <- false;
	    path.dist <- 1;
	    c_pathn = path;				// Start of other end
	    built = true;
	    break;
	  }

	  local tile = path.tile;
	  local ptile = ppath.tile;

	  dist = AIMap.DistanceManhattan(tile,ptile);	// To next tile

	  path.dist <- dist;
	  c_dist += dist;

	  if((AIBridge.IsBridgeTile(tile) || AITunnel.IsTunnelTile(tile)) &&
	     (AIBridge.IsBridgeTile(ptile) || AITunnel.IsTunnelTile(ptile)))
	    path.exists <- true;
	  else {
	    path.exists <- false;

	    if(dist > 1) {
	      Error("Missing pass");
	      c_error = YATT_E_INTERNAL;
	      return false;
	    }

	    local done = false;

	    while (!done && !AIRoad.BuildRoad(tile,ptile)) {
	      switch(c_error = AIError.GetLastError()) {
	      case AIError.ERR_ALREADY_BUILT:
		path.exists = true;
		done = true;
		break;

	      case AIError.ERR_VEHICLE_IN_THE_WAY:
		switch (HandleVehicleInTheWay(tile,ptile)) {
		case YATT_E_SUCCESS:
		  done = true;
		  break;

		case YATT_E_TRAFFIC:
		  Warning("Too much traffic at: "+tile+" "+ptile);HandleVehicleInTheWay
		  return false;				// Too much traffic
		}

		break;					// Retry

	      case AIError.ERR_AREA_NOT_CLEAR:
		if(!DemolishTile(ptile)) return false;

		done = true;
		break;

	      default:
		Error("Test build failed: "+AIError.GetLastErrorString());
		return false;
	      }
	    }
	  }

	  npath = path;
	  path = ppath;
	  path.next <- npath;
	}

	if(built) {
	  c_cost = acc.GetCosts();

	  c_ends.reverse();				// 0 is now src end
	  c_front.reverse();
	}

	c_path = ConvertToArray();	// So path is not nested and can be saved
	c_pathn = null;

	return built;
}
	function cATTRoad::ExtendRoad(idx,tile)
//	---------------------------------------
//	Extend road at one end by adding the tile, cost not updated
//	Must be called after TestBuildRoad as it needs array of tables
//	idx		Index to either end of road {0=next links|1}
//	tile		tile to add to road
{
	c_front[idx] = c_ends[idx];		// Prev end is now front
	c_ends[idx] = tile;			// New end tile
	c_dist++;				// One tile longer

	local	path = {tile = tile};		// New path table

	path.exists <- false;
	path.length <- 0;
	path.cost <- 0;
	path.dist <- 1;

	if(!idx)				// Start of path
	  c_path.insert(0,path);
	else					// End of path
	  c_path.push(path);
}
	function cATTRoad::BuildRoad(endidx=1,tiles=null)
//	-------------------------------------------------
//	Builds the road
//	endidx		0/1 - Start at this end
//	tiles		null or number of tiles to build
//	Returns		true if road built ok
{
	local	idx,limit=c_path.len()-1,path,link,n=0;

	if(!endidx) {					// Follow next links
	  idx = 0;
	  link = 1;
	}
	else {						// Follow prev links
	  idx = limit;
	  link = -1;
	}

	path = c_path[idx];

	while (path != null) {
	  idx += link;

	  local npath = (idx >= 0 && idx <= limit) ? c_path[idx] : null;

	  if(npath == null || (tiles && tiles == n)) break;

	  if(!path.exists) {
	    local tile = path.tile;
	    local ntile = npath.tile;

	    if(path.dist == 1) {
	      local done = false;

	      while (!done && !AIRoad.BuildRoad(tile,ntile)) {
	        switch (c_error = AIError.GetLastError()) {
	        case AIError.ERR_ALREADY_BUILT:
		  done = true;
		  break;

		case AIError.ERR_VEHICLE_IN_THE_WAY:
		  Sleep(10);
		  break;

		case AIError.ERR_NOT_ENOUGH_CASH:
		  WaitForFunds();
		  break;

		case AIError.ERR_AREA_NOT_CLEAR:
		  if(!DemolishTile(ntile)) return false;

//		  In case the problem is with tile, not ntile

		  if(!AIRoad.BuildRoad(tile,ntile)) return false;

		  done = true;
		  break;

		default:
		  Error("Error building road: "+AIError.GetLastErrorString());
		  return false;
		}
	      }
	    }
	    else {
	      Error("Need to build pass");
	      c_error = YATT_E_INTERNAL;
	      return false;
	    }
	  }

	  path = npath;
	  n++;
	}

	return true;
}
	function cATTRoad::SurveyRoadDepot(endidx,ignore=null)
//	------------------------------------------------------
//	Looks at tiles perpendicular to road for places for depots
//	Must be called after TestBuild as it needs array of tables
//	endidx		Index to either end of road {0=next links|1}
//	ignore		Array of tiles to ignore (roadloop tiles)
//	Returns		null or a table with the following fields:
//			depot		Existing depot object
//			tile		Depot tile
//			front		Depot front tile
{
	local	idx,limit=c_path.len()-4,path,npath,link,found=0,dtile=null,tile;
	local	test = AITestMode();

	if(!endidx) {					// Follow next links
	  idx = 3;
	  link = 1;
	}
	else {						// Follow prev links
	  idx = limit;
	  link = -1;
	}

	local	stile = c_ends[endidx];			// Station tile
	local	etile = c_ends[1-endidx];		// Other end

	npath = c_path[idx];

	for (; idx >= 3 && idx <= limit; idx += link) {
	  path = npath;
	  npath = c_path[idx + link];

	  local	i,dir = path.dir;			// Road direction
	  tile = path.tile;

	  if(AIBridge.IsBridgeTile(tile) || AITunnel.IsTunnelTile(tile) ||
	     AITile.IsStationTile(tile)) continue;	// Through stop

	  if(BuiltSlope(tile,npath.tile) & ATT_SLP_SLOPED) continue;

	  local elev = AITile.GetMaxHeight(tile);

	  while (true) {
	    for (i=0; i <= 1; i++) {			// Try 2 perp dirs
	      local pdir = s_mapdirp[dir][i];
	      dtile = tile + s_mapoffs[pdir];		// Perp tile L/R
	      local pelev = AITile.GetMaxHeight(dtile);

	      if(elev != pelev) continue;
	      if(AITile.IsCoastTile(dtile)) continue;
	      if(AIMap.DistanceManhattan(dtile,stile) <= 3) continue;// Too close
	      if(AIMap.DistanceManhattan(dtile,etile) <= 3) continue;// Too close

//	      Make sure we don't pick a tile where this road will be built

	      if(dtile == c_path[idx-1].tile || dtile == c_path[idx+1].tile) continue;

	      if(ignore) {				// Have ignore tiles
		local skip = false;

		foreach (j,itile in ignore) {
		  if(dtile == itile) {
		    skip = true;
		    break;
		  }
		}

		if(skip) continue;			// Skip location
	      }

	      if(AIRoad.IsRoadDepotTile(dtile) &&
		 AICompany.IsMine(AITile.GetOwner(dtile)) &&
		 AIRoad.AreRoadTilesConnected(dtile,tile)) {
		found = 2;				// Existing depot
		break;
	      }

	      if(AIRoad.BuildRoadDepot(dtile,tile)) {	// Test build
	        found = 1;				// New depot
	        break;
	      }
	    }

//	    If the current tile direction does not match the next tile direction
//	    the road turned. Check for possible spot on other corner too.

	    if(found || dir == npath.dir) break;

	    dir = npath.dir;
	  }

	  if(found || AIMap.DistanceManhattan(tile,etile) <= 2) break;
	}

	test = null;

	if(!found) return null;

	local	ret = {tile = dtile, front = tile};

	ret.depot <- (found == 1) ? null : s_g.depots[dtile];

	return ret;
}
	function cATTRoad::GetUnbuiltTiles()
//	------------------------------------
//	Scan path and return array with tiles that need road built
//	Must be called after TestBuildRoad as it needs array of tables
{
	local	idx,limit=c_path.len()-1,path;
	local	arr = [];

	idx = limit;
	path = c_path[idx];

	while (path != null) {
	  if(!path.exists || idx == limit || !c_path[idx+1].exists) arr.push(path.tile);

	  path = (idx) ? c_path[--idx] : null;
	}

	return (arr.len()) ? arr : null;
}	
	function cATTRoad::ConvertToArray()
//	-----------------------------------
//	Convert the nested c_path tables to an array of tables
{
	local	next,arr = [];
	local	path = c_pathn;

	while (path != null) {
	  next = path.next;

	  arr.push(path);

	  delete path.next;
	  delete path.prev;

	  path = next;
	}

	return arr;
}
